import xarray as xr

@xr.register_dataarray_accessor("traces")

class traces(object):
    '''
    This is a classe to work on sismic traces using xarray.
    '''
    
    def __init__(self, xarray_obj):
        '''
        Constructor for traces. 
        
        :param xarray_obj:
        :type xarray_obj: xr.DataArray
        '''
        self._obj = xarray_obj 
    pass

from sisicepy.traces.io import datacube_loader
from sisicepy.traces.recorder import datacube
from sisicepy.traces.signal import freq_filter
from sisicepy.traces.signal import utils
from sisicepy.traces.signal import processing
from sisicepy.traces.captor import geobit
from sisicepy.traces.captor import digos



import xarray as xr
import pyrocko.io
import numpy as np
from datetime import datetime, timezone, timedelta
import pandas as pd
import pint_xarray
import natsort
import glob
from pint_xarray import unit_registry as ureg

def load_datacube(adr,axis=[0]):
    '''
    Loader for DIGOS DataCube

    :param adr: Path to the DataCube file
    :type adr: str
    :param axis: List of axes to load
    :type axis: list of int
    :return: xarray Dataset containing the loaded data and metadata
    :rtype: xarray.Dataset
    '''

    data=pyrocko.io.load(adr,format='datacube')


    df=xr.Dataset()
    
    for i in axis:
        df['axis_'+str(i)]=xr.DataArray(data[i].ydata,dims='time')*ureg.bit
        for name in data[0].meta:
            df['axis_'+str(i)].attrs[name]=data[0].meta[name]
            df['axis_'+str(i)].attrs[name]=data[0].meta[name]
        
    ti=datetime.fromtimestamp(data[0].tmin,tz=timezone.utc)
    tf=datetime.fromtimestamp(data[0].tmax,tz=timezone.utc)
    df['time']=pd.date_range(ti,tf,data[0].data_len()).values
    
    
    for name in data[0].meta:
        df.attrs[name]=data[0].meta[name]
        df.attrs[name]=data[0].meta[name]
    
    return df

def load_mfdatacube(adr_list,**kwargs):
    '''
    Loader for DIGOS DataCube

    :param adr_list:
    :type adr_list: list
    '''

    data_list=[]
    for adr in adr_list:
        data_list.append(load_datacube(adr,**kwargs))

    return xr.concat(data_list,dim='time')

def open_day_datacube(path,day,**kwargs):
    '''
    Loader for DIGOS DataCube for a given day

    :param day: format %Y-%m-%d
    :type axis: str
    '''
    files=natsort.natsorted(glob.glob(path))
    # function to parse the path from datacube to string date
    def extract_date(file):
        fs = file.split('/')
        return '20' + fs[-2][0:2] + '-' + fs[-1][0:2] + '-' + fs[-1][2:4]

    formatted_dates = list(map(extract_date, files))

    ###########################################################
    # find file for the given day and one day before and afer #
    ###########################################################
    formatted_dates_dt = np.array([datetime.strptime(date, '%Y-%m-%d') for date in formatted_dates])

    # Define the target date (D-Day)
    target_date = datetime.strptime(day, '%Y-%m-%d')

    # Define the previous day and next day
    previous_day = target_date - timedelta(days=1)
    next_day = target_date + timedelta(days=1)

    # Use np.where to find indices for the target date, previous day, and next day
    indices = np.where(
        (formatted_dates_dt == target_date) |
        (formatted_dates_dt == previous_day) |
        (formatted_dates_dt == next_day)
    )

    #############
    # Open data #
    #############
    if len(indices[0]) != 0:
        di=day+'T00:00:00'
        df=next_day.strftime('%Y-%m-%d')+'T00:00:00'
        return load_mfdatacube(np.array(files)[indices[0]]).sel(time=slice(di,df),**kwargs)
    else:
        print('No data at this date !')




import xarray as xr
import numpy as np
import pint_xarray
from pint import Quantity
from pint_xarray import unit_registry as ureg

def trapezoid(self,band,delta):
    '''
    Trapeziod band filter:
    .. line-block::  
        ........Lower band...........Higher band........
        ........band[0]...............band[1]...........
        ...........|------------------|.................
        ..........||..................||................
        .........|||..................|||...............
        ........||||..................||||..............
        .......|||||..................|||||.............
        ......||||||..................||||||............
        _____|||||||..................|||||||_______....
        .......delta..................delta.............


    :param self: signal to filter with trapeziode band
    :type self: xr.DataArray
    :param band: lower band and higher band ex: [4.5,90] Hz
    :type band: list
    :param delta: delta for the begining and the end of the trapez
    :type delta: float
    '''

    da=self._obj

    # delta t 
    dt=np.int32(da.time[1]-da.time[0])*10**-9 # because time unit is ns

    # compute FFT of da
    da_FFT=np.fft.fftshift(np.fft.fft(da))
    # and associated frequencies
    da_freq=np.fft.fftshift(np.fft.fftfreq(len(da),d=dt))
    
    ######################
    ## Construct trapez ##
    ######################
    # trap_side is a function to extract value for the linear part
    # between 0 and 1 sign=+1 and 1 and 0 sign=0
    def trap_side_pos(x,delta,Pup,sign=1):
        return sign*1/delta*x+1-sign*Pup/delta
    
    def trap_side_neg(x,delta,Pup,sign=1):
        return sign*1/delta*x+1+sign*Pup/delta
    
    # Build trapez for >0 frequencies in the spectrum
    # Build 0 1 array with 1 value in the band interval
    Top_Trap_U=(da_freq>band[0])*(da_freq<band[1])
    # Build line part 0 to 1 
    Up_Trap_U=(da_freq>=band[0]-delta)*(da_freq<=band[0])
    Up_Trap_U=trap_side_pos(da_freq,delta,band[0],sign=1)*Up_Trap_U
    # Build line part 1 to 0
    Dw_Trap_U=(da_freq>=band[1])*(da_freq<=band[1]+delta)
    Dw_Trap_U=trap_side_pos(da_freq,delta,band[1],sign=-1)*Dw_Trap_U

    # Build trapez for >0 frequencies in the spectrum
    # Build 0 1 array with 1 value in the band interval
    Top_Trap_B=(da_freq<-1*band[0])*(da_freq>-1*band[1])
    # Build line part 0 to 1 
    Up_Trap_B=(da_freq>=-1*band[1]-delta)*(da_freq<=-1*band[1])
    Up_Trap_B=trap_side_neg(da_freq,delta,band[1],sign=1)*Up_Trap_B
    # Build line part 1 to 0
    Dw_Trap_B=(da_freq>=-1*band[0])*(da_freq<=-1*band[0]+delta)
    Dw_Trap_B=trap_side_neg(da_freq,delta,band[0],sign=-1)*Dw_Trap_B


    # compute full trapez
    all_trap=Top_Trap_U+Up_Trap_U+Dw_Trap_U+Top_Trap_B+Up_Trap_B+Dw_Trap_B
    trapez=xr.DataArray(all_trap,dims=['freq'])
    trapez['freq']=da_freq
    
    ###############################
    ## Apply full trapez to data ##
    ###############################
    # Apply only on the real part
    FFT_filtered=da_FFT*trapez*ureg.second*da.pint.units

    # Compute filtered signal
    signal_filtered=xr.DataArray(np.fft.ifft(np.fft.ifftshift(FFT_filtered)),dims=da.dims)*da.pint.units
    for name in da.dims:
        signal_filtered[name]=da[name]

    return signal_filtered.real,FFT_filtered,trapez


def poles_zeros_filter(self,ZPG,waterlevel):
    '''
    Pole and zeros filter
    (https://en.wikipedia.org/wiki/Pole%E2%80%93zero_plot)

    :param self: signal to filter
    :type self: xr.DataArray
    :param ZPG: [zeros,poles,gains]
    :type ZPG: list
    :param waterlevel:
    :type waterlevel: float
    '''
    
    da=self._obj

    # delta t 
    dt=np.int32(da.time[1]-da.time[0])*10**-9 # because time unit is ns

    # compute FFT of da
    da_FFT=np.fft.fftshift(np.fft.fft(da))
    # and associated frequencies
    da_freq=np.fft.fftshift(np.fft.fftfreq(len(da),d=dt))

    ####################################
    ## Construct poles_zeros_function ##
    ####################################
    def poles_zeros(zeros,poles,gains,freq,waterlevel):
        N=1
        for Zi in zeros:
            N = N*(2*np.pi*1j*freq - Zi)

        D = 1
        for Pi in poles:
            D = D*(2*np.pi*1j*freq - Pi)
        
        transfer=gains*N/D
        # waterlevel
        temp1=transfer*np.conj(transfer)
        gamma=np.max(temp1)*waterlevel
        filter_np=np.conj(transfer)/(temp1+gamma)
        filter=xr.DataArray(filter_np,dims='freq')
        filter['freq']=freq

        
        return filter
    
    filter_poles_zeros=poles_zeros(ZPG[0],ZPG[1],ZPG[2],da_freq,waterlevel)
    ###############################
    ## Apply full trapez to data ##
    ###############################
    # Apply only on the real part
    FFT_filtered=da_FFT*filter_poles_zeros*ureg.second*da.pint.units

    # Compute filtered signal
    signal_filtered=xr.DataArray(np.fft.ifft(np.fft.ifftshift(FFT_filtered)),dims=da.dims)*FFT_filtered.pint.units/ureg.second
    
    for name in da.dims:
        signal_filtered[name]=da[name]
     
    return signal_filtered.real,FFT_filtered,filter_poles_zeros

#----------------------------------------------------------------------------------------------
xr.DataArray.traces.trapezoid = trapezoid
xr.DataArray.traces.poles_zeros_filter = poles_zeros_filter



import xarray as xr
import numpy as np
import pint_xarray
from pint_xarray import unit_registry as ureg

def geophone_ZPK():
    '''
    Value for Poles and Zeros function of the Digos Geophone

    Zeros: [0,0]
    Poles :[-19.78+20.20i,-19.78-20.20i]
    K=27.7 V/(m.s^-1)
    '''

    Zs=[0,0]
    Ps=[np.complex128(-19.78+20.20j),np.complex128(-19.78-20.20j)]
    K = 27.7*ureg.volt*ureg.second/ureg.meter

    return Zs,Ps,K


import xarray as xr
import scipy.signal
import pint_xarray
from pint_xarray import unit_registry as ureg

def remove_mean(self):
    '''
    Remove mean from da signal

    :param self: signal
    :type self: xr.DataArray
    '''
    
    return self._obj-self._obj.mean()

def remove_trend(self):
    '''
    Remove linear trend from da signal

    :param self: signal
    :type self: xr.DataArray
    '''

    da=self._obj

    da_no_trend=xr.DataArray(scipy.signal.detrend(da),dims=da.dims)
    for name in da.dims:
        da_no_trend[name]=da[name]

    return da_no_trend*da.pint.units

#---------------------------------------------------------------------------
xr.DataArray.traces.remove_mean = remove_mean
xr.DataArray.traces.remove_trend = remove_trend


import xarray as xr
import numpy as np
import pint_xarray
from pint_xarray import unit_registry as ureg

def count2volt_datacube(self,Vpp=4.096,nb_bin=2**24):
    '''
    Convert numeric value to Voltage for DataCube

    :param Vpp: peak to peak value, default: 4.096 (unit V)
    :type Vpp: float
    :param nb_bin: number of integer for the numerisation, default: 2**24
    :type nb_bin: int
    '''

    coeff   =  Vpp/nb_bin/np.int32(self._obj.attrs['P_AMPL']) * ureg.volt/ureg.bit
    
    return self._obj*coeff

#-----------------------------------------------------------------------------
xr.DataArray.traces.count2volt_datacube = count2volt_datacube


import xarray as xr
import numpy as np
import pint_xarray
from pint import Quantity
from pint_xarray import unit_registry as ureg
import scipy.signal

def PSD(self,**kwargs):
    '''
    Wrapper of `scipy.signal.welch` function for Power Spectral Density.
    :param self:
    :type self: xr.DataArray
    '''

    # sampling frequency
    fs=1/(np.float64((self._obj.time[1]-self._obj.time[0]))*10**-9)

    freq,psd=scipy.signal.welch(self._obj,fs=fs,**kwargs)
    
    da_psd=xr.DataArray(psd,dims='freq')*(self._obj.pint.units)**2/ureg.hertz
    da_psd['freq']=freq

    return da_psd

def spectrogram(self,win,t_start,dB=True,**kwargs):
    '''
     Wrapper of `scipy.signal.welch` to performed PSD spectrogram
    :param self:
    :type self: xr.DataArray
    :param win: time length in seconds
    :type win: int
    :param t_start: data for the start
    :type t_start: str
    :param dB: value in dB
    :type dB: bool
    '''
    da=self._obj
    fs=1/(np.float64((self._obj.time[1]-self._obj.time[0]))*10**-9)

    # Sub data selection
    da_sel=da.sel(time=slice(t_start,None))
    # Find the number of win interval in the time serie
    nb_per=int(len(da_sel)/(win*fs))
    # select the right time length for the spectrogram
    da_sel=da_sel[0:np.int32(win*fs*nb_per)]

    freq,psd=scipy.signal.welch(da_sel.values.reshape(nb_per,int(win*fs)),fs,**kwargs)
    # Covert in dB
    if dB:
        psd=10*np.log10(psd)
        unit_dB=ureg.dB
    else:
        unit_dB=1.

    da_psd=xr.DataArray(np.transpose(psd),dims=['freq','time'])*(self._obj.pint.units)**2/ureg.hertz
    da_psd['freq']=freq

    tr_time=da_sel.time.values.reshape(nb_per,int(win*fs))
    da_psd['time']=tr_time[:,0]

    return da_psd


#----------------------------------------------------------------------------------------------
xr.DataArray.traces.PSD = PSD
xr.DataArray.traces.spectrogram = spectrogram


import xarray as xr
import numpy as np
import pint_xarray
from pint_xarray import unit_registry as ureg


def C100_ZPK():
    '''
    Value for Poles and Zeros function of the Geobit C100

    Zeros: [0+0i,0+0i]
    Poles :[-15.8336-23.4251i,-15.8336+23.4251i]
    K = 1 V/(m.s^-1)
    '''

    Zs=[0+0j,0+0j]
    Ps=[np.complex128(-15.8336-23.4251j),np.complex128(-15.8336+23.4251j)]
    K = 27.7*ureg.volt*ureg.second/ureg.meter

    return Zs,Ps,K

def S100_ZPK():
    '''
    Value for Poles and Zeros function of the Geobit S100

    Zeros: [0+0i,0+0i,0+0i]
    Poles :[-615+0i,-0.406-0.606i,-0.406+0.606i,-1.226313+0j]
    K = 1 V/(m.s^-1)
    '''

    Zs=[0+0j,0+0j,0+0j]
    Ps=[np.complex128(-615+0j),np.complex128(-0.406-0.606j),np.complex128(-0.406+0.606j),np.complex128(-1.226313+0j)]
    K = 616*ureg.volt*ureg.second/ureg.meter

    return Zs,Ps,K